/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 1994, 1995 Waldorf Electronics
 * Written by Ralf Baechle and Andreas Busse
 * Copyright (C) 1995 - 1999 Ralf Baechle
 * Copyright (C) 1996 Paul M. Antoine
 * Modified for DECStation and hence R3000 support by Paul M. Antoine
 * Further modifications by David S. Miller and Harald Koerfgen
 * Copyright (C) 1999 Silicon Graphics, Inc.
 *
 * Kevin Kissell, kevink@mips.com and Carsten Langgaard, carstenl@mips.com
 * Copyright (C) 2000 MIPS Technologies, Inc.  All rights reserved.
 */
#include <linux/config.h>
#include <linux/init.h>
#include <linux/threads.h>

#include <asm/asm.h>
#include <asm/current.h>

#ifndef CONFIG_KBUILD_2_5
#include <asm/offset.h>
#else
#include "asm-offsets.h"
#endif

#include <asm/pgtable-bits.h>
#include <asm/processor.h>
#include <asm/regdef.h>
#include <asm/cachectl.h>
#include <asm/mipsregs.h>
#include <asm/stackframe.h>

		.text
#ifndef CONFIG_TIVO
		/*
		 * Reserved space for exception handlers.
		 * Necessary for machines which link their kernels at KSEG0.
		 */
		.fill	0x400
#endif

		/* The following two symbols are used for kernel profiling. */
		EXPORT(stext)
		EXPORT(_stext)

		__INIT

		/* Cache Error */
		LEAF(except_vec2_generic)
		.set	noreorder
		.set	noat
		.set    mips0
		/*
		 * This is a very bad place to be.  Our cache error
		 * detection has triggered.  If we have write-back data
		 * in the cache, we may not be able to recover.  As a
		 * first-order desperate measure, turn off KSEG0 cacheing.
		 */
		mfc0	k0,CP0_CONFIG
		li	k1,~CONF_CM_CMASK
		and	k0,k0,k1
		ori	k0,k0,CONF_CM_UNCACHED
		mtc0	k0,CP0_CONFIG
		/* Give it a few cycles to sink in... */
		nop
		nop
		nop

		j	cache_parity_error
		nop
		END(except_vec2_generic)

		.set	at

		/*
		 * Special interrupt vector for embedded MIPS.  This is a
		 * dedicated interrupt vector which reduces interrupt processing
		 * overhead.  The jump instruction will be inserted here at
		 * initialization time.  This handler may only be 8 bytes in
		 * size!
		 */
		NESTED(except_vec4, 0, sp)
1:		j	1b			/* Dummy, will be replaced */
		 nop
		END(except_vec4)

		/*
		 * EJTAG debug exception handler.
		 * The EJTAG debug exception entry point is 0xbfc00480, which
		 * normally is in the boot PROM, so the boot PROM must do a
		 * unconditional jump to this vector.
		 */
		NESTED(except_vec_ejtag_debug, 0, sp)
		j	ejtag_debug_handler
		 nop
		END(except_vec_ejtag_debug)

		/*
		 * EJTAG debug exception handler.
		 */
		NESTED(ejtag_debug_handler, PT_SIZE, sp)
		.set	noat
		.set	noreorder
		mtc0	k0, CP0_DESAVE
		mfc0	k0, CP0_DEBUG

		sll	k0, k0, 30	# Check for SDBBP.
		bgez	k0, ejtag_return

		la	k0, ejtag_debug_buffer
		sw	k1, 0(k0)
		SAVE_ALL
		jal	ejtag_exception_handler
		 move	a0, sp
		RESTORE_ALL
		la	k0, ejtag_debug_buffer
		lw	k1, 0(k0)

ejtag_return:
		mfc0	k0, CP0_DESAVE
		.word	0x4200001f     # DERET, return from EJTAG debug exception.
		 nop
		.set	at
		END(ejtag_debug_handler)

		/*
		* NMI debug exception handler for MIPS reference boards.
		* The NMI debug exception entry point is 0xbfc00000, which
		* normally is in the boot PROM, so the boot PROM must do a
		* unconditional jump to this vector.
		*/
		NESTED(except_vec_nmi, 0, sp)
		j       nmi_handler
		 nop
		END(except_vec_nmi)

		NESTED(nmi_handler, PT_SIZE, sp)
		.set    noat
		.set    noreorder
		.set    mips3
		SAVE_ALL
		jal     nmi_exception_handler
		 move   a0, sp
		RESTORE_ALL
		eret
		.set    at
		.set    mips0
		END(nmi_handler)

		/*
		 * Kernel entry point
		 */
		NESTED(kernel_entry, 16, sp)
		.set	noreorder

#if defined(CONFIG_TIVO) && defined(CONFIG_BLK_DEV_INITRD)
                /*
                 * Sigh, what a pain... the ramdisk image overlays .bss, so
                 * need to copy it off before clearing .bss
                 */
                la	t0, 0x5469566f
                bne	t0, a2, 2f
                 la	t1, _end
                addiu	t1, 0x0fff
                ori	t1, 0x0fff
                xori	t1, 0x0fff
                lw	t3, 20(a3)
                lw	t4, 24(a3)
                lw	t2, 4(a3)
                andi	t0, t3, 3
                bnez	t0, 2f
                 addi	t4, 4
                ori	t4, 3
                xori	t4, 3
                subu	t5, t4, t3
                blez	t5, 2f
                 addu	t3, t2, t4
                addu	t6, t1, t5
1:		lw	t0, -4(t3)
                addiu	t6, -4
                sw	t0, (t6)
                bne	t6, t1, 1b
                 addiu	t3, -4
2:
#endif

		/* The firmware/bootloader passes argc/argp/envp
		 * to us as arguments.  But clear bss first because
		 * the romvec and other important info is stored there
		 * by prom_init().
		 */
		la	t0, _edata
		sw	zero, (t0)
		la	t1, (_end - 4)
1:
		addiu	t0, 4
		bne	t0, t1, 1b
		 sw	zero, (t0)

		/*
		 * Stack for kernel and init, current variable
		 */
		la	$28, init_task_union

#ifdef CONFIG_4KSTACK
		sw	$28, kernels28
		nop

		addiu	t0, $28, (2*KERNEL_STACK_SIZE)-32
		subu	sp, t0, 4*SZREG

		sw	t0, kernelsp
                sw      t0, TASK_KSTACK($28)
		nop
		la	t0, isr_stack
	
		/* 8K stack */
		addiu	t0, t0, (_PAGE_SIZE << 1)-32

		subu	t0, t0, 4*SZREG
		sw	t0, intrsp
#else
		addiu	t0, $28, (KERNEL_STACK_SIZE)-32
		subu	sp, t0, 4*SZREG

		sw	t0, kernelsp
#endif

		jal	init_arch
		 nop
		END(kernel_entry)


#ifdef CONFIG_SMP

/*
 * SMP slave cpus entry point.  Board specific code for bootstrap calls this
 * function after setting up the stack and gp registers.
 */
		LEAF(smp_bootstrap)
		.set push
		.set noreorder
		mtc0	zero, CP0_WIRED
		CLI
		mfc0	t0, CP0_STATUS
		li	t1, ~(ST0_CU1|ST0_CU2|ST0_CU3|ST0_KX|ST0_SX)
		and	t0, t1
		or	t0, (ST0_CU0);
		jal	start_secondary
		mtc0	t0, CP0_STATUS
		.set pop
		END(smp_bootstrap)
#endif

		__FINIT

		/*
		 * This buffer is reserved for the use of the EJTAG debug
		 * handler.
		 */
		.data
		EXPORT(ejtag_debug_buffer)
		.fill	4

		.comm	pgd_current, NR_CPUS * 8, 8
		.comm	kernelsp,    NR_CPUS * 8, 8
#ifdef CONFIG_4KSTACK
		.comm	kernels28,   NR_CPUS * 8, 8
		.comm	intrsp,	     4
		.comm	intrctr,     4 
#endif

	.macro	page name, order=0
	.globl	\name
\name:	.size	\name, (_PAGE_SIZE << \order)
	.align	12	
	.fill	(_PAGE_SIZE << \order),1,0
	.type	\name, @object
	.endm

	.section .data.page_aligned,"aw"

	page	swapper_pg_dir, _PGD_ORDER
	page	empty_bad_page, 0
	page	empty_bad_page_table, 0
	page	invalid_pte_table, 0
#ifdef CONFIG_4KSTACK
	/* 8k stack */
	page	isr_stack,	1
#endif
