/*
 * Copyright (C) 2006 TiVo Inc. All Rights Reserved.
 */ 
#ifndef FASTBUFS_H
#define FASTBUFS_H   

#include<linux/ioctl.h>
#include <linux/tivo-monotime.h>
#ifdef __KERNEL__
#include <asm/time.h>
#else
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif
/*ioctl*/
#define FBUF_IOC_MAGIC          'f'

#define FBUF_ALLOC              _IO(FBUF_IOC_MAGIC, 1)
#define FBUF_FREE               _IO(FBUF_IOC_MAGIC, 2)

#define FBUF_MAGIC_VMSTART      0x40000000
#define FBUF_MAGIC_WORD         "F A S T B U F F E R S\0"

typedef enum {
    FBUF_THREAD_EXEC_ENTRY=0,
    FBUF_CNTXT_ENTRY,
    FBUF_MAX_TYPES
}fbuf_type_t;

struct context_entry{
    char name[60];
    void *cs_user;
};

struct thread_exec_entry{
    pid_t pid;
    short reserved;
    struct context_entry *cntxt;
    void *tls_user[6];
};

struct fastbuf_hdr{
    unsigned char magic_word[64];
    monotonic_t t; 
    struct  timeval tv;
    pid_t cur_pid;
    unsigned long rt_pri;
    struct thread_exec_entry *cur_exec_entry;
    volatile unsigned int rseq;
};

struct fbuf_req{
    fbuf_type_t fbuf_type;
    void *ptr;  
};
static struct fastbuf_hdr *fastbuf = (struct fastbuf_hdr *)FBUF_MAGIC_VMSTART;

#ifndef __KERNEL__
/*
 *Fastbuffers are owned by kernel and theoretically 
 *user tasks should have read-only access to these
 *buffers. On MIPS while sharing buffers across both
 *domains by manupulating TLBs, exclusive read-only
 *protection for user domain cannot be acheived. Hence for 
 *safety reasons always access fastbuffer elements using the
 *following API.
 *
 */ 
static int is_fastbuf_intialized(void){
    return ( fastbuf && !strcmp(fastbuf->magic_word, FBUF_MAGIC_WORD));
}
static pid_t fast_getpid(void){
    return fastbuf->cur_pid;
}
static unsigned long fast_getrtpri(void){
    return fastbuf->rt_pri;
}
static struct thread_exec_entry * fast_get_curthread_execinfo(void){
    return fastbuf->cur_exec_entry;
}
static struct context_entry *fast_get_thread_cntxt(void){
    if ( fastbuf->cur_exec_entry )
        return fastbuf->cur_exec_entry->cntxt;
}
static void update_curthread_cntxt(struct context_entry *cntxt){
    if ( fastbuf->cur_exec_entry )
        fastbuf->cur_exec_entry->cntxt = cntxt;
}

static void fast_gettimeofday(struct timeval *tv){
    unsigned int temp;
    do {
        temp = fastbuf->rseq;
        tv->tv_sec = fastbuf->tv.tv_sec;
        tv->tv_usec = fastbuf->tv.tv_usec;
    }while(temp != fastbuf->rseq );
}

static void fast_getmonotime(monotonic_t *t){
    unsigned int temp;
    do {
        temp = fastbuf->rseq;
        t->count = fastbuf->t.count;
    }while(temp != fastbuf->rseq );
}

static void *fbuf_alloc(fbuf_type_t fbuf_type){
        
    int fd;
    struct fbuf_req fbuf_req;

    fd = open("/dev/fbuf",O_RDONLY);
    if ( fd < 0 )
        return NULL;
    fbuf_req.fbuf_type = fbuf_type;
    if ( ioctl(fd,FBUF_ALLOC,&fbuf_req))
        fbuf_req.ptr = NULL;
    close(fd);

    return fbuf_req.ptr;
}

static void fbuf_free(void *ptr){
    int fd;

    fd = open("/dev/fbuf",O_RDONLY);
    if ( fd < 0 )
        return;
    if ( !ioctl(fd,FBUF_FREE,ptr))
        ptr = NULL;

    close(fd);
}

#endif /*ifndef __KERNEL*/

#endif /* FASTBUFS_H */
